# (C) British Crown Copyright 2015, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus._Elementwise`."""

from itertools import permutations
from functools import partial
import unittest

import numpy as np
import numpy.ma as ma
from numpy.testing import assert_array_equal

from biggus import _Elementwise as Elementwise


class Test__masked_arrays(unittest.TestCase):
    def setUp(self):
        a = np.arange(6).reshape(3, 2) - 3
        self.mask = a % 3 == 0
        self.masked_array = np.ma.masked_array(a, self.mask)

    def test_single_argument_operation(self):
        expected = np.abs(self.masked_array)
        actual = Elementwise(self.masked_array, None, np.abs, ma.abs)
        result = actual.masked_array()
        assert_array_equal(result.mask, self.mask)
        assert_array_equal(result, expected)

    def test_dual_argument_operation(self):
        exponent = 2
        expected = self.masked_array ** exponent
        actual = Elementwise(self.masked_array, exponent, np.power, ma.power)
        result = actual.masked_array()
        assert_array_equal(result.mask, self.mask)
        assert_array_equal(result, expected)

    def test_no_masked_array_function(self):
        result = Elementwise(self.masked_array, None, np.sign)
        msg = 'No sign operation defined for masked arrays'
        with self.assertRaisesRegexp(TypeError, msg):
            result.masked_array()


class Test_dual_argument_dtype(unittest.TestCase):
    def setUp(self):
        dtypes = []
        for dtype in ['int', 'uint', 'float']:
            dtypes.extend(np.sctypes[dtype])
        self.pairs = [pair for pair in permutations(dtypes, 2)]

    def common_dtype(self, dtype1, dtype2, op):
        return op(np.ones(1, dtype=dtype1),
                  np.ones(1, dtype=dtype2)).dtype

    def _test(self, *ops):
        for dtype1, dtype2 in self.pairs:
            a1 = np.ones(1, dtype=dtype1)
            a2 = np.ones(1, dtype=dtype2)
            actual = Elementwise(a1, a2, *ops)
            expected_dtype = self.common_dtype(dtype1, dtype2, ops[0])
            self.assertEqual(actual.dtype, expected_dtype)

    def test_add(self):
        self._test(np.add, ma.add)

    def test_sub(self):
        self._test(np.subtract, ma.subtract)

    def test_multiply(self):
        self._test(np.multiply, ma.multiply)

    def test_floor_divide(self):
        self._test(np.floor_divide, ma.floor_divide)

    def test_true_divide(self):
        self._test(np.true_divide, ma.true_divide)

    def test_power(self):
        self._test(np.power, ma.power)


class Test_single_argument_dtype(unittest.TestCase):
    def setUp(self):
        dtypes = []
        for dtype in ['int', 'uint', 'float']:
            dtypes.extend(np.sctypes[dtype])
        self.dtypes = dtypes

    def dtype_after_of(self, dtype, op):
        return op(np.ones(1, dtype=dtype)).dtype

    def _test(self, *ops):
        for dtype in self.dtypes:
            a1 = np.ones(1, dtype=dtype)
            actual = Elementwise(a1, None, *ops)
            expected_dtype = self.dtype_after_of(dtype, ops[0])
            self.assertEqual(actual.dtype, expected_dtype)

    def test_abs(self):
        self._test(np.abs, ma.abs)

    def test_rint(self):
        self._test(np.rint)

    def test_sign(self):
        self._test(np.sign)

    def test_square(self):
        self._test(np.square)

    def test_partial_function(self):
        clip_between = partial(np.clip, a_min=1, a_max=4)
        self._test(clip_between)

        array = np.arange(10000, dtype=np.float32)
        actual = Elementwise(array, None, clip_between)
        assert_array_equal(actual.ndarray(), clip_between(array))
        self.assertEqual(actual.ndarray().max(), 4)


class Test_single_argument(unittest.TestCase):
    def test_getitem(self):
        a1 = np.arange(3)
        actual = Elementwise(a1, None, np.cos)
        self.assertEqual(actual[1].ndarray(), np.cos(a1[1]))


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2015, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Integration tests for maths operations."""

import numpy as np
import numpy.ma as ma
import unittest

import biggus
import biggus.tests.unit._aggregation_test_framework as test_framework


class TestSum(unittest.TestCase):

    def test_sum_float(self):
        a = biggus.ConstantArray((1), dtype=np.float32)
        b = a + 1
        self.assertEqual('float32', b.dtype)

    def test_sum_int(self):
        a = biggus.ConstantArray((1), dtype=np.int32)
        b = a + 1
        self.assertEqual('int32', b.dtype)


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2014 - 2015, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.

from contextlib import contextmanager

try:
    from unittest import mock
except ImportError:
    import mock

import numpy as np

import biggus


class AccessCounter(object):
    """
    Something that acts like a NumPy ndarray, but which records how
    many times each element has been read.

    """
    def __init__(self, ndarray):
        self._ndarray = ndarray
        self.counts = np.zeros(ndarray.shape)

    @property
    def dtype(self):
        return self._ndarray.dtype

    @property
    def ndim(self):
        return self._ndarray.ndim

    @property
    def shape(self):
        return self._ndarray.shape

    def __array__(self):
        return self._ndarray

    def __getitem__(self, keys):
        self.counts[keys] += 1
        return self._ndarray[keys]

    def unique_counts(self):
        return set(np.unique(self.counts))


class _KeyGen(object):
    """Gives you the content of keys when indexing."""
    def __getitem__(self, keys):
        return keys

#: An object that can be indexed to return a usable key.
key_gen = _KeyGen()


@contextmanager
def set_chunk_size(value):
    old_chunk_size = biggus.MAX_CHUNK_SIZE
    biggus.MAX_CHUNK_SIZE = value
    yield
    biggus.MAX_CHUNK_SIZE = old_chunk_size



# (C) British Crown Copyright 2014 - 2015, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for `biggus._all_slices`."""

import unittest

import numpy as np

import biggus
from biggus import _all_slices, _all_slices_inner
from biggus.tests import set_chunk_size


class Test__all_slices(unittest.TestCase):

    def test_all_cases(self):
        array = biggus.ConstantArray((4, 3, 5), dtype=np.float32)
        # Chunk size set to fit in two items from the second dimension into a
        # single chunk, but not the whole dimension.
        chunk_size = 5 * 3 - 1
        with set_chunk_size(chunk_size):
            slices = _all_slices(array)
        expected = [[0, 1, 2, 3],
                    [slice(0, 2, None), slice(2, 3, None)],
                    (slice(None, None, None),)]
        self.assertEqual(slices, expected)

    def test_always_slices(self):
        array = biggus.ConstantArray((3, 5), dtype=np.float32)
        chunk_size = 5 - 1
        with set_chunk_size(chunk_size):
            slices = _all_slices_inner(array.shape, always_slices=True)
        expected = [[slice(0, 1, None), slice(1, 2, None), slice(2, 3, None)],
                    [slice(0, 4, None), slice(4, 5, None)]]
        self.assertEqual(slices, expected)


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2014 - 2015, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""Integration tests for aggregations."""

import unittest

import numpy as np

import biggus
from biggus.tests import set_chunk_size


class Test(unittest.TestCase):
    def test_var_masked(self):
        data = biggus.ConstantArray((10, 400, 720), dtype=np.float32)
        v = biggus.var(data, axis=1)
        result = v.masked_array()
        self.assertIsInstance(result, np.ma.MaskedArray)
        self.assertEqual(result.shape, (10, 720))
        self.assertTrue(np.all(result == 0))
        self.assertTrue(np.all(result.mask == 0))


class TestAggregation__mixed_dtypes(unittest.TestCase):
    def test(self):
        shape = (3, 5, 10)
        a = biggus.ConstantArray(shape, dtype=np.float32)
        b = biggus.ConstantArray(shape, dtype=np.float64)

        with set_chunk_size(32//8*10-1):
            result = biggus.sum(a * b, axis=0).ndarray()
            self.assertEqual(result.shape, shape[1:])


class Test__slices_with_mathematical_filter(unittest.TestCase):

    def setUp(self):
        self.dtype = np.float32

    def _biggus_filter(self, data, weights):
        # Filter a data array (time, <other dimensions>) using information in
        # weights dictionary.
        #
        # Args:
        #
        # * data:
        #     biggus array of the data to be filtered
        # * weights:
        #     dictionary of absolute record offset : weight

        # Build filter_matrix (time to time' mapping).
        shape = data.shape

        # Build filter matrix as a numpy array and then populate.
        filter_matrix_np = np.zeros((shape[0], shape[0])).astype(self.dtype)

        for offset, value in weights.items():
            filter_matrix_np += np.diag([value] * (shape[0] - offset),
                                        k=offset)
            if offset > 0:
                filter_matrix_np += np.diag([value] * (shape[0] - offset),
                                            k=-offset)

        # Create biggus array for filter matrix, adding in other dimensions.
        for _ in shape[1:]:
            filter_matrix_np = filter_matrix_np[..., np.newaxis]

        filter_matrix_bg_single = biggus.NumpyArrayAdapter(filter_matrix_np)

        # Broadcast to correct shape (time, time', lat, lon).
        filter_matrix_bg = biggus.BroadcastArray(
            filter_matrix_bg_single, {i+2: j for i, j in enumerate(shape[1:])})

        # Broadcast filter to same shape.
        biggus_data_for_filter = biggus.BroadcastArray(data[np.newaxis, ...],
                                                       {0: shape[0]})

        # Multiply two arrays together and sum over second time dimension.
        filtered_data = biggus.sum(biggus_data_for_filter * filter_matrix_bg,
                                   axis=1)

        # Cut off records at start and end of output array where the filter
        # cannot be fully applied.
        filter_halfwidth = len(weights) - 1
        filtered_data = filtered_data[filter_halfwidth:-filter_halfwidth]

        return filtered_data

    def test__biggus_filter(self):
        shape = (1451, 1, 1)

        # Generate dummy data as biggus array.
        numpy_data = np.random.random(shape).astype(self.dtype)
        biggus_data = biggus.NumpyArrayAdapter(numpy_data)

        # Information for filter...
        # Dictionary of weights: key = offset (absolute value), value = weight
        weights = {0: 0.4, 1: 0.2, 2: 0.1}
        # This is equivalent to a weights array of [0.1, 0.2, 0.4, 0.2, 0.1].
        filter_halfwidth = len(weights) - 1

        # Filter data
        filtered_biggus_data = self._biggus_filter(biggus_data, weights)

        # Extract eddy component (original data - filtered data).
        eddy_biggus_data = (biggus_data[filter_halfwidth:-filter_halfwidth] -
                            filtered_biggus_data)

        # Aggregate over time dimension.
        mean_eddy_biggus_data = biggus.mean(eddy_biggus_data, axis=0)

        # Force evaluation.
        mean_eddy_numpy_data = mean_eddy_biggus_data.ndarray()

        # Confirm correct shape.
        self.assertEqual(mean_eddy_numpy_data.shape, shape[1:])


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2015, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""
Unit tests for `biggus._ufunc_wrapper` and the functions that it
has wrapped.

"""

import inspect
import unittest
import warnings

import numpy as np
import numpy.ma
from numpy.testing import assert_array_equal
from numpy.ma.testutils import assert_array_equal as assert_masked_array_equal

import biggus


class Test__unary_fn_wrapper(unittest.TestCase):
    def test_nin2_docstring(self):
        wrapped_fn = biggus._ufunc_wrapper(np.add)
        doc = inspect.getdoc(wrapped_fn)

        expected = ('Return the elementwise evaluation of '
                    'np.add(a, b) as another Array.')
        self.assertEqual(doc, expected)

    def test_nin1_docstring(self):
        wrapped_fn = biggus._ufunc_wrapper(np.negative)
        doc = inspect.getdoc(wrapped_fn)

        expected = ('Return the elementwise evaluation of '
                    'np.negative(a) as another Array.')
        self.assertEqual(doc, expected)

    def test_non_ufunc(self):
        msg = 'not a ufunc'
        with self.assertRaisesRegexp(TypeError, msg):
            biggus._ufunc_wrapper(lambda x: x)

    def test_nout2_ufunc(self):
        msg = "Unsupported ufunc 'modf' with 1 input arrays & 2 output arrays."
        with self.assertRaisesRegexp(ValueError, msg):
            biggus._ufunc_wrapper(np.modf)


class Test_wrapped_functions(unittest.TestCase):
    def setUp(self):
        self.arr1 = np.array([1, 2, 3])
        self.arr2 = np.array([2, 1, 2])
        self.biggus_arr1 = biggus.NumpyArrayAdapter(self.arr1)
        self.biggus_arr2 = biggus.NumpyArrayAdapter(self.arr2)

        self.marr1 = np.ma.masked_array([1, 2, 3], mask=[0, 1, 0])
        self.marr2 = np.ma.masked_array([1, 5, 2], mask=[0, 0, 1])
        self.biggus_marr1 = biggus.NumpyArrayAdapter(self.marr1)
        self.biggus_marr2 = biggus.NumpyArrayAdapter(self.marr2)

        ufunc_names = ['absolute', 'add', 'arccos', 'arccosh', 'arcsin',
                       'arcsinh', 'arctan', 'arctan2', 'arctanh',
                       'bitwise_and', 'bitwise_or', 'bitwise_xor',
                       'ceil', 'conj', 'cos', 'cosh', 'deg2rad', 'divide',
                       'equal', 'exp', 'exp2', 'expm1', 'floor',
                       'floor_divide', 'fmax', 'fmin', 'greater',
                       'greater_equal', 'hypot', 'invert', 'left_shift',
                       'less', 'less_equal', 'log',
                       'log10', 'log2', 'logical_and', 'logical_not',
                       'logical_or', 'logical_xor', 'maximum', 'minimum',
                       'multiply', 'negative', 'not_equal', 'power',
                       'rad2deg', 'reciprocal', 'right_shift', 'rint', 'sign',
                       'sin', 'sinh', 'sqrt', 'square', 'subtract', 'tan',
                       'tanh', 'true_divide', 'trunc']

        self.ufuncs = [(name, getattr(np, name), getattr(biggus, name))
                       for name in ufunc_names]

    def test_all_fns(self):
        self.ufuncs.append(('sub', np.subtract, biggus.subtract))

        for fn_name, ufunc, biggus_fn in self.ufuncs:
            if ufunc.nin == 1:
                result = biggus_fn(self.biggus_arr1)
                with warnings.catch_warnings(record=True) as expected_warnings:
                    warnings.simplefilter("always")
                    expected = ufunc(self.arr1)
            else:
                result = biggus_fn(self.biggus_arr1, self.biggus_arr2)
                with warnings.catch_warnings(record=True) as expected_warnings:
                    warnings.simplefilter("always")
                    expected = ufunc(self.arr1, self.arr2)

            with warnings.catch_warnings(record=True) as actual_warnings:
                warnings.simplefilter("always")
                actual = result.ndarray()

            self.assertEqual([unicode(warning.message)
                              for warning in actual_warnings],
                             [unicode(warning.message)
                              for warning in expected_warnings])

            error_msg = ('biggus.{} produces different results to np.{}:'
                         ''.format(fn_name, fn_name))
            assert_array_equal(actual, expected, error_msg)

    def test_all_fns_masked(self):
        for fn_name, ufunc, biggus_fn in self.ufuncs:
            masked_ufunc = getattr(np.ma, ufunc.__name__, None)
            if masked_ufunc is not None:
                if ufunc.nin == 1:
                    result = biggus_fn(self.biggus_marr1)
                    with warnings.catch_warnings(record=True):
                        warnings.simplefilter("always")
                        expected = masked_ufunc(self.marr1)
                else:
                    result = biggus_fn(self.biggus_marr1, self.biggus_marr2)
                    with warnings.catch_warnings(record=True):
                        warnings.simplefilter("always")
                        expected = masked_ufunc(self.marr1, self.marr2)

                with warnings.catch_warnings(record=True):
                    warnings.simplefilter("always")
                    actual = result.masked_array()

                error_msg = ('biggus.{} produces different results to '
                             'np.ma.{}'.format(fn_name, ufunc.__name__))
                assert_masked_array_equal(actual, expected, error_msg)


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2015, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""
Unit tests for `biggus._unary_fn_wrapper` and the functions that it
has wrapped.

"""

import inspect
import sys
import unittest

import numpy as np
from numpy.testing import assert_array_equal

import biggus


class Test__unary_fn_wrapper(unittest.TestCase):
    # Attach a future proof assert raises method.
    if sys.version_info[0] == 2:
        assertRaisesRegex = unittest.TestCase.assertRaisesRegexp

    def test_docstring(self):
        wrapped_fn = biggus._unary_fn_wrapper('my_module.my_function',
                                              lambda a: a)
        doc = inspect.getdoc(wrapped_fn)

        expected = ('Return the elementwise evaluation of '
                    'my_module.my_function(a) as another Array.')
        self.assertEqual(doc, expected)

    def test_auto_fn_name(self):
        wrapped_fn = biggus._unary_fn_wrapper('my_module.my_function',
                                              lambda a: a)
        self.assertEqual(wrapped_fn.__name__, (lambda a: a).__name__)

    def test_given_fn_name(self):
        wrapped_fn = biggus._unary_fn_wrapper('my_module.my_function',
                                              lambda a: a, fn_name='identity')
        self.assertEqual(wrapped_fn.__name__, 'identity')

    def test_masked_array_creates_elementwise(self):
        wrapped_fn = biggus._unary_fn_wrapper('my_module.my_function',
                                              lambda a: a + 10,
                                              lambda a: a - 10,
                                              fn_name='identity')
        self.assertIsInstance(wrapped_fn(np.array([1])), biggus._Elementwise)

    def test_unexpected_n_arguments(self):
        wrapped_fn = biggus._unary_fn_wrapper('my_module.my_function',
                                              lambda a: a + 10,
                                              fn_name='identity')
        # TODO: It would be good if this were not called "wrapped_function".
        if sys.version_info[0] == 2:
            msg = 'wrapped_function\(\) takes exactly 1 argument \(2 given\)'
        else:
            msg = ('wrapped_function\(\) takes 1 positional argument but 2 '
                   'were given')
        with self.assertRaisesRegex(TypeError, msg) as ex:
            wrapped_fn(np.array([-5, 2]), np.array([-5, 2]))

    def test_ndarray_expected_values(self):
        wrapped_fn = biggus._unary_fn_wrapper('my_module.my_function',
                                              lambda a: a + 10,
                                              fn_name='identity')
        assert_array_equal(wrapped_fn(np.array([-5, 2])).ndarray(),
                           [5, 12])

    def test_masked_array_expected_values(self):
        wrapped_fn = biggus._unary_fn_wrapper('my_module.my_function',
                                              lambda a: a + 10,
                                              lambda a: a - 10,
                                              fn_name='identity')
        assert_array_equal(wrapped_fn(np.array([-5, 2])).masked_array(),
                           [-15, -8])

    def test_masked_array_undefined(self):
        wrapped_fn = biggus._unary_fn_wrapper('my_module.my_function',
                                              lambda a: a + 10,
                                              fn_name='identity')
        msg = 'No <lambda> operation defined for masked arrays.'
        with self.assertRaisesRegex(TypeError, msg):
            assert_array_equal(wrapped_fn(np.array([-5, 2])).masked_array(),
                               [-15, -8])


class Test_wrapped_functions(unittest.TestCase):
    def test_all_fns(self):
        fns_to_test = ['isreal', 'iscomplex', 'isinf', 'isnan', 'signbit']
        arr = np.array([-10, 0, 5])
        biggus_arr = biggus.NumpyArrayAdapter(arr)
        for fn_name in fns_to_test:
            np_fn = getattr(np, fn_name)
            biggus_fn = getattr(biggus, fn_name)
            result = biggus_fn(biggus_arr)
            assert_array_equal(result.ndarray(), np_fn(arr))


if __name__ == '__main__':
    unittest.main()



# (C) British Crown Copyright 2015, Met Office
#
# This file is part of Biggus.
#
# Biggus is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Biggus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Biggus. If not, see <http://www.gnu.org/licenses/>.
"""
Unit tests for `biggus._dual_input_fn_wrapper` and the functions that it
has wrapped.

"""

import inspect
import unittest

import numpy as np
from numpy.testing import assert_array_equal

import biggus


class Test__dual_input_fn_wrapper(unittest.TestCase):
    def test_docstring(self):
        wrapped_fn = biggus._dual_input_fn_wrapper('my_module.my_function',
                                                   lambda a, b: a + b)
        doc = inspect.getdoc(wrapped_fn)

        expected = ('Return the elementwise evaluation of '
                    'my_module.my_function(a, b) as another Array.')
        self.assertEqual(doc, expected)

    def test_auto_fn_name(self):
        wrapped_fn = biggus._dual_input_fn_wrapper('my_module.my_function',
                                                   lambda a, b: a + b)
        self.assertEqual(wrapped_fn.__name__, (lambda a, b: a + b).__name__)

    def test_given_fn_name(self):
        wrapped_fn = biggus._dual_input_fn_wrapper('my_module.my_function',
                                                   lambda a, b: a + b,
                                                   fn_name='identity')
        self.assertEqual(wrapped_fn.__name__, 'identity')

    def test_masked_array_creates_elementwise(self):
        wrapped_fn = biggus._dual_input_fn_wrapper('my_module.my_function',
                                                   lambda a, b: a + b,
                                                   lambda a, b: a - b,
                                                   fn_name='identity')
        result = wrapped_fn(np.array([-5, 2]), np.array([-5, 2]))
        self.assertIsInstance(result, biggus._Elementwise)

    def test_ndarray_expected_values(self):
        wrapped_fn = biggus._dual_input_fn_wrapper('my_module.my_function',
                                                   lambda a, b: a + b,
                                                   fn_name='identity')
        result = wrapped_fn(np.array([-5, 2]), np.array([-5, 2]))
        assert_array_equal(result.ndarray(), [-10, 4])

    def test_masked_array_expected_values(self):
        wrapped_fn = biggus._dual_input_fn_wrapper('my_module.my_function',
                                                   lambda a, b: a - b,
                                                   lambda a, b: a + b,
                                                   fn_name='identity')
        result = wrapped_fn(np.array([-5, 2]), np.array([-5, 2]))
        assert_array_equal(result.masked_array(), [-10, 4])

    def test_masked_array_undefined(self):
        wrapped_fn = biggus._dual_input_fn_wrapper('my_module.my_function',
                                                   lambda a, b: a + b,
                                                   fn_name='identity')
        msg = 'No <lambda> operation defined for masked arrays.'
        result = wrapped_fn(np.array([-5, 2]), np.array([-5, 2]))
        with self.assertRaisesRegexp(TypeError, msg):
            assert_array_equal(result.masked_array(), [0, 4])


class Test_wrapped_functions(unittest.TestCase):
    def test_all_fns(self):
        fns_to_test = ['copysign', 'nextafter', 'ldexp', 'fmod']
        arr1 = np.array([10, 2, 5])
        arr2 = np.array([1, 5, 15])
        biggus_arr1 = biggus.NumpyArrayAdapter(arr1)
        biggus_arr2 = biggus.NumpyArrayAdapter(arr2)

        for fn_name in fns_to_test:
            np_fn = getattr(np, fn_name)
            biggus_fn = getattr(biggus, fn_name)
            result = biggus_fn(biggus_arr1, biggus_arr2)
            assert_array_equal(result.ndarray(), np_fn(arr1, arr2))


if __name__ == '__main__':
    unittest.main()
